@@ -1,7 +1,6 @@
 import logging
-from dataclasses import dataclass, fields
+from dataclasses import dataclass
 from datetime import timedelta
-from itertools import chain
 from typing import Any, Callable, Coroutine, Optional, Protocol, TypedDict, cast
 
 from telebot import AsyncTeleBot
@@ -16,7 +15,10 @@ from telebot_components.feedback.anti_spam import (
     AntiSpamInterface,
     AntiSpamStatus,
 )
-from telebot_components.feedback.trello_integration import TrelloIntegration
+from telebot_components.feedback.trello_integration import (
+    MessageRepliedFromTrelloContext,
+    TrelloIntegration,
+)
 from telebot_components.redis_utils.interface import RedisInterface
 from telebot_components.stores.banned_users import BannedUsersStore
 from telebot_components.stores.category import CategoryStore
@@ -31,6 +33,7 @@ from telebot_components.stores.language import (
     Language,
     LanguageStore,
     any_text_to_str,
+    vaildate_singlelang_text,
 )
 
 
@@ -106,10 +109,10 @@ class FeedbackConfig:
 class FeedbackHandler:
     """
     A class incapsulating the following workflow:
-     - people write messages to a bot
-     - bot forwards them to admin chat
-     - admins reply to messages
-     - bot copy messages back to the user
+     - people write messages to the bot
+     - the bot forwards messages to admin chat
+     - admins reply
+     - the bot copies messages back to the user
     """
 
     def __init__(
@@ -220,15 +223,13 @@ class FeedbackHandler:
         if self.config.force_category_selection and self.service_messages.you_must_select_category is None:
             raise ValueError("force_category_selection is True, you_must_select_category message must be set")
 
-        if self.language_store is not None:
-            languages = self.language_store.languages
-        else:
-            languages = [None]
         for message in self.service_messages.user_facing:
             if message is None:
                 continue
-            for language in languages:
-                any_text_to_str(message, language)
+            if self.language_store is None:
+                vaildate_singlelang_text(message)
+            else:
+                self.language_store.validate_multilang(message)
 
     async def _ban_admin_chat_action(
         self, admin_message: tg.Message, forwarded_message: tg.Message, origin_chat_id: int
@@ -243,15 +244,15 @@ class FeedbackHandler:
         await self.message_log_store.push(origin_chat_id, forwarded_message.id, reset_ttl=True)
 
     def _admin_help_message(self) -> str:
-        help_msg = (
-            "<b>Справка-памятка для админского чата</b>\n\n"
-            + "💬 <i>Основное</i>\n"
+        paragraphs = ["<b>Справка-памятка для админского чата</b>"]
+        paragraphs.append(
+            "💬 <i>Основное</i>\n"
             + "· Сюда бот пересылает все сообщения (кроме специальных случаев вроде /команд), которые ему "
             + "пишут в личку.\n"
-            + "· Если в этом чате ответить на сообщение, бот скопирует ответ в чат с пользовател_ьницей.\n"
+            + "· Если в этом чате ответить на сообщение, бот скопирует ответ в чат с пользовател_ьницей."
         )
         if self.category_store is not None:
-            help_msg += (
+            categories_help = (
                 "📊 <i>Категории сообщений</i>\n"
                 + "· Каждо_й пользовател_ьнице предлагается выбрать одну из категорий: "
                 + ", ".join(
@@ -260,30 +261,35 @@ class FeedbackHandler:
                 + "\n"
             )
             if self.config.force_category_selection:
-                help_msg += "· Выбор категории обязателен для пользовател_ьниц.\n"
+                categories_help += "· Выбор категории обязателен для пользовател_ьниц."
             else:
-                help_msg += "· Выбор категории необязателен, боту можно написать и без него.\n"
+                categories_help += "· Выбор категории необязателен, боту можно написать и без него."
 
-        help_msg += (
+            paragraphs.append(categories_help)
+
+        security_help = (
             "🛡️ <i>Защита и безопасность</i>\n"
             + "· Бот никак не выдаёт, кто отвечает пользовател_ьнице из этого чата. Насколько возможно судить, "
             + "никакого способа взломать бота нет. Однако всё, что вы отвечаете через бота, сразу пересылается человеку "
-            + "на другом конце, и на данный момент удалить отправленное сообщение нельзя, поэтому будьте внимательны!\n"
+            + "на другом конце, и на данный момент удалить отправленное сообщение нельзя, поэтому будьте внимательны!"
         )
         if isinstance(self.anti_spam, AntiSpam):
-            help_msg += (
-                "· Бот автоматически ограничивает число сообщений, прислыаемых ему в единицу времени. "
+            security_help += (
+                "\n"
+                + "· Бот автоматически ограничивает число сообщений, прислыаемых ему в единицу времени. "
                 + f"Конфигруация на данный момент: не больше {self.anti_spam.config.throttle_after_messages} сообщений за "
-                + f"{self.anti_spam.config.throttle_duration}; при необходимости её можно изменять.\n"
+                + f"{self.anti_spam.config.throttle_duration}. При необходимости её можно изменять."
             )
         if self.banned_users_store is not None:
-            help_msg += (
-                "· Если ответить на пересланное сообщение командой /ban, "
-                + "пользовател_ьница будет заблокирован_а, а все сообщения от них в чате — удалены\n\n"
+            security_help += (
+                "\n· Если ответить на пересланное сообщение командой /ban, "
+                + "пользовател_ьница будет заблокирован_а, а все сообщения от них в чате — удалены"
             )
-        help_msg += (
+        paragraphs.append(security_help)
+
+        paragraphs.append(
             "📋 <i>История сообщений</i>\n"
-            + "· Через бота может быть неудобно вести несколько длительных переписок — все они мешаются в одном чате."
+            + "· Через бота может быть неудобно вести несколько длительных переписок — все они мешаются в одном чате.\n"
             + "· Если ответить на пересланное сообщение командой /log, бот перешлёт всю историю переписки с "
             + "пользовател_ьницей "
             + (
@@ -295,33 +301,27 @@ class FeedbackHandler:
             )
         )
         if self.trello_integration is not None:
-            help_msg += "\n\n🗂️ <i>Интеграция с Trello</i>\n"
-            help_msg += (
-                f"· Помимо чата сообщения выгружаются на <a href={self.trello_integration.board.url}>доску</a> Trello"
+            trello_help = "🗂️ <i>Интеграция с Trello</i>\n"
+            trello_help += (
+                f'· Помимо чата сообщения выгружаются на <a href="{self.trello_integration.board.url}">доску</a> Trello '
+                + f"в списки: "
+                + ", ".join(f"<b>{l.name}</b>" for l in self.trello_integration.lists_by_category_name.values())
+                + "\n"
+                "· В карточки переносятся присланные в бот фотографии и документы. Для каждого сообщения доступна "
+                + "обратная ссылка на этот чат."
             )
-            if self.trello_integration.categories is not None:
-                help_msg += "в несколько списков по категориям (хештегам): "
-                help_msg += ", ".join(
-                    f"<b>{l.name}</b>" for l in self.trello_integration.lists_by_category_name.values()
+            if self.trello_integration.reply_with_card_comments:
+                trello_help += (
+                    "\n"
+                    + "· Через комментарии к карточке можно ответить на сообщение в боте: "
+                    + "комментарий «/reply текст ответа» значит, что бот отправит «текст ответа» в чат с "
+                    + "пользовател_ьницей, а также напишет уведомление сюда."
                 )
-            else:
-                help_msg += f"в список <b>{self.trello_integration.bot_prefix}</b>"
-            help_msg += (
-                ". "
-                + "В карточки переносятся присланные в бот фотографии и документы; для каждого сообщения доступна "
-                + "обратная ссылка на этот чат.\n"
-                # + "· В тестовом режиме работает ответ через Trello: если прокоментировать карточку сообщением в "
-                # + "формате «/reply текст ответа», бот отправит «текст ответа» в чат с пользовател_ьницей, а также "
-                # + "напишет уведомление сюда."
-            )
+            paragraphs.append(trello_help)
 
         if self.config.admin_chat_help_extra:
-            help_msg += "\n\n🪄 <i>Другое</i>\n"
-            help_msg += self.config.admin_chat_help_extra
-
-        # TODO: add bot support hotline link :)
-
-        return help_msg
+            paragraphs.append("🪄 <i>Другое</i>\n" + self.config.admin_chat_help_extra)
+        return "\n\n".join(paragraphs)
 
     async def _user_message_filter(self, message: tg.Message) -> bool:
         if self.config.custom_user_message_filter is not None:
@@ -338,13 +338,12 @@ class FeedbackHandler:
         user: tg.User,
         message_forwarder: Callable[[], Coroutine[None, None, tg.Message]],
         user_replier: Callable[[str, Optional[tg.ReplyMarkup]], Coroutine[None, None, Any]],
-        origin_message_to_export: Optional[tg.Message],
-    ):
+    ) -> Optional[tg.Message]:
         if self.banned_users_store is not None and await self.banned_users_store.is_banned(user.id):
-            return
+            return None
         anti_spam_status = await self.anti_spam.status(user)
         if anti_spam_status is AntiSpamStatus.SOFT_BAN:
-            return
+            return None
 
         if self.language_store is not None:
             language = await self.language_store.get_user_language(user)
@@ -353,7 +352,7 @@ class FeedbackHandler:
         if anti_spam_status is AntiSpamStatus.THROTTLING:
             anti_spam = cast(AntiSpam, self.anti_spam)  # only real AntiSpam can return this status
             await user_replier(self.service_messages.throttling(anti_spam.config, language), None)
-            return
+            return None
 
         if self.config.hashtags_in_admin_chat:
             category_hashtag = None  # sentinel
@@ -367,7 +366,7 @@ class FeedbackHandler:
                             any_text_to_str(you_must_select_category, language),
                             await self.category_store.markup_for_user(user),
                         )
-                        return
+                        return None
                 else:
                     category_hashtag = category.hashtag
 
@@ -420,12 +419,6 @@ class FeedbackHandler:
                     await self.recently_sent_confirmation_flag_store.set_flag(user.id)
 
         if self.trello_integration is not None:
-            if origin_message_to_export is None:
-                self.logger.warning(
-                    "Trello integration is set up, but user message handler "
-                    + "was not supplied with original message to export to Trello"
-                )
-                return
             category = await self.category_store.get_user_category(user) if self.category_store else None
 
             def postprocess_card_description(descr: str) -> str:
@@ -436,31 +429,36 @@ class FeedbackHandler:
                     descr = descr + "\n\n" + "[post]:" + postforwarded_msg.text_content
                 return descr
 
-            await self.trello_integration.export_message(
-                origin_message_to_export,
-                forwarded_msg,
-                category,
+            await self.trello_integration.export_user_message(
+                user=user,
+                forwarded_message=forwarded_msg,
+                category=category,
                 postprocess_card_description=postprocess_card_description,
             )
+        return forwarded_msg
 
-    async def emulate_user_message(self, bot: AsyncTeleBot, user: tg.User, text: str, **send_message_kwargs):
+    async def emulate_user_message(
+        self, bot: AsyncTeleBot, user: tg.User, text: str, no_response: bool = False, **send_message_kwargs
+    ) -> Optional[tg.Message]:
         """Sometimes we want FeedbackHandler to act like the user has sent us a message, but without actually
         a message there (they might have pressed a button or interacted with the bot in some other way). This
-        message can be used in such cases.
+        method can be used in such cases.
         """
 
         async def message_forwarder() -> tg.Message:
             return await bot.send_message(self.admin_chat_id, text=text, **send_message_kwargs)
 
-        async def user_replier(text: str, reply_markup: Optional[tg.ReplyMarkup]) -> tg.Message:
-            return await bot.send_message(user.id, text=text, reply_markup=reply_markup)
+        async def user_replier(text: str, reply_markup: Optional[tg.ReplyMarkup]) -> Optional[tg.Message]:
+            if no_response:
+                return None
+            else:
+                return await bot.send_message(user.id, text=text, reply_markup=reply_markup)
 
-        await self._handle_user_message(
+        return await self._handle_user_message(
             bot=bot,
             user=user,
             message_forwarder=message_forwarder,
             user_replier=user_replier,
-            origin_message_to_export=None,
         )
 
     def setup(self, bot: AsyncTeleBot):
@@ -484,7 +482,6 @@ class FeedbackHandler:
                 user=message.from_user,
                 message_forwarder=message_forwarder,
                 user_replier=user_replier,
-                origin_message_to_export=message,
             )
 
         @bot.message_handler(chat_id=[self.admin_chat_id], commands=["help"])
@@ -525,6 +522,14 @@ class FeedbackHandler:
             finally:
                 await self.hashtag_message_for_forwarded_message_store.save(message_id, hashtag_message_data)
 
+        async def on_message_replied_from_trello(context: MessageRepliedFromTrelloContext):
+            await self.message_log_store.push(context.origin_chat_id, context.reply_message_id)
+            if self.config.hashtags_in_admin_chat:
+                await _remove_unanswered_hashtag(context.forwarded_user_message_id)
+
+        if self.trello_integration is not None:
+            self.trello_integration.set_on_message_replied_from_trello(on_message_replied_from_trello)
+
         @bot.message_handler(
             chat_id=[self.admin_chat_id],
             is_reply=True,
@@ -585,7 +590,7 @@ class FeedbackHandler:
                             + ", ".join(available_commands),
                         )
                 else:
-                    # actual response to user
+                    # actual response to the user
                     await bot.copy_message(
                         chat_id=origin_chat_id, from_chat_id=self.admin_chat_id, message_id=message.id
                     )
@@ -595,6 +600,8 @@ class FeedbackHandler:
                         await bot.reply_to(message, self.service_messages.copied_to_user_ok)
                     if self.config.hashtags_in_admin_chat:
                         await _remove_unanswered_hashtag(forwarded_msg.id)
+                    if self.trello_integration is not None:
+                        await self.trello_integration.export_admin_message(message, to_user_id=origin_chat_id)
             except Exception as e:
                 await bot.reply_to(message, f"Something went wrong: {e}")
                 self.logger.exception(f"Unexpected error while replying to forwarded msg")
